/**
 * @copyright Copyright 2023-2024 RnD Center "ELVEES", JSC
 */

#include <dirent.h>
#include <stdio.h>
#include <sys/file.h>
#include <sys/stat.h>

#include <glib.h>
#include <gst/gst.h>
#include <gst/rtsp-server/rtsp-server.h>

#define DEFAULT_SRC                                                            \
  "videotestsrc ! video/x-raw,format=NV12,width=1280,height=720 ! "            \
  "interpipesink name=src"
#define DEFAULT_SINKS                                                          \
  "{"                                                                          \
  "\"/main\": \"interpipesrc stream-sync=restart-ts is-live=true "             \
  "listen-to=src ! queue "                                                     \
  "! x264enc speed-preset=ultrafast tune=zerolatency ! rtph264pay "            \
  "name=pay0\", "                                                              \
  "\"/secondary\": \"interpipesrc stream-sync=restart-ts is-live=true "        \
  "listen-to=src ! queue ! videoscale "                                        \
  "! video/x-raw,format=NV12,width=640,height=360 ! queue ! jpegenc ! "        \
  "rtpjpegpay name=pay0\""                                                     \
  "}"
#define DEFAULT_RTSP_PORT "8554"
#define DEFAULT_DUMP_FLX_DIR "/tmp"
#define DEFAULT_DUMP_FLX_CONTEXT 0
#define DEFAULT_DUMP_FLX_DISABLE_TIMESTAMP FALSE

#define READY_FLAG_PATH READY_FLAG_DIR "/" READY_FLAG_FILE

#define MULTI_LAUNCH_SUMMARY                                                   \
  "Utility runs RTSP server with multiple streams."                            \
  "When felixsrc/felixmultisrc source elements are used in pipeline,"          \
  "an image in a flx format can be saved on SIGUSR1"                           \
  "(see --dump-flx-* options)."

#define MULTI_LAUNCH_DESCRIPTION                                               \
  "External applications should read " READY_FLAG_PATH " file to ensure that " \
  "an image was taken and get a name of an image from the file."

#ifdef DEBUG
/* There is more convenient to send SIGINT via the Ctrl+C combination during
 * the debug.
 */
#define DUMP_FLX_SIGNAL SIGINT
#else
#define DUMP_FLX_SIGNAL SIGUSR1
#endif /* DEBUG */

static char *_src = DEFAULT_SRC;
static char *_sinks = DEFAULT_SINKS;
static char *_port = DEFAULT_RTSP_PORT;
static char *_dump_dir = DEFAULT_DUMP_FLX_DIR;
static int _dump_context = DEFAULT_DUMP_FLX_CONTEXT;
static char *_dump_src_type = NULL;
static gboolean _dump_disable_ts = DEFAULT_DUMP_FLX_DISABLE_TIMESTAMP;

static GOptionEntry entries[] = {
    {"src", 's', 0, G_OPTION_ARG_STRING, &_src,
     "Source pipeline (default: " DEFAULT_SRC ")", "SOURCE"},
    {"sinks", 'k', 0, G_OPTION_ARG_STRING, &_sinks,
     "Dict in JSON format {\"RTSP path\": \"RTSP bin\"} "
     "(default: " DEFAULT_SINKS ")",
     "SINKS"},
    {"port", 'p', 0, G_OPTION_ARG_STRING, &_port,
     "Port to listen on (default: " DEFAULT_RTSP_PORT ")", "PORT"},
    {"dump-flx-directory", 0, 0, G_OPTION_ARG_FILENAME, &_dump_dir,
     "A directory for storing images (default: '" DEFAULT_DUMP_FLX_DIR "')",
     "DUMP_FLX_DIR"},
    {"dump-flx-context", 0, 0, G_OPTION_ARG_INT, &_dump_context,
     "Felix context number from which images are saved "
     "(default: " G_STRINGIFY(
         DEFAULT_DUMP_FLX_CONTEXT) "). Supported values: 0, 1.",
     "DUMP_FLX_CONTEXT"},
    {"dump-flx-type", 0, 0, G_OPTION_ARG_STRING, &_dump_src_type,
     "Type of image. Supported values: YUV, Display, Bayer.", "DUMP_FLX_TYPE"},
    {"dump-flx-disable-timestamp", 0, 0, G_OPTION_ARG_NONE, &_dump_disable_ts,
     "Disable timestamp in image name "
     "(default: " G_STRINGIFY(DEFAULT_DUMP_FLX_DISABLE_TIMESTAMP) ")",
     "DUMP_FLX_DISABLE_TIMESTAMP"},
    {NULL}};

static GstElement *felixsrc = NULL;
static char *image_path;
static char *image_lock;
int lock_fd;

static GstElement *get_felixsrc_element(GstElement *pipeline) {
  const char *types[] = {"GstFelixsrc", "GstFelixMultiSrc"};

  if (!pipeline) {
    g_printerr("Pipeline is NULL\n");
    return NULL;
  }

  for (int i = 0; i < sizeof(types) / sizeof(types[0]); i++) {
    GType type = g_type_from_name(types[i]);
    if (!type) {
      continue;
    }
    GstElement *element = gst_bin_get_by_interface(GST_BIN(pipeline), type);
    if (!element) {
      continue;
    }
    return element;
  }

  return NULL;
}

static void image_ready_event(GstElement *element, gboolean image_success,
                              gpointer user_data) {
  if (!image_success) {
    g_printerr("Failed to dump image\n");
    goto exit;
  }

  g_print("Image status will be saved to '" READY_FLAG_PATH "'\n");
  FILE *fp = fopen(READY_FLAG_PATH, "w");
  if (!fp) {
    g_printerr("Failed to open file " READY_FLAG_PATH "\n");
    goto exit;
  }

  fprintf(fp, "Image is ready: %s\n", image_path);
  fclose(fp);

exit:
  g_free(image_path);
  if (close(lock_fd)) {
    g_printerr("Failed to close %s file\n", image_lock);
  }
  if (unlink(image_lock)) {
    g_printerr("Failed to remove %s file\n", image_lock);
  }
}

static void dump_to_flx() {
  DIR *dir = opendir(READY_FLAG_DIR);
  if (dir) {
    closedir(dir);
  } else if (mkdir(READY_FLAG_DIR, 1777)) {
    g_printerr("Failed to create directory " READY_FLAG_DIR "\n");
    return;
  }

  image_lock = READY_FLAG_PATH ".lock";
  lock_fd = open(image_lock, O_CREAT, 0444);
  if (lock_fd < 0) {
    g_printerr("Failed to open %s file\n", image_lock);
    return;
  }

  if (flock(lock_fd, LOCK_EX)) {
    g_printerr("Failed to lock %s file\n", image_lock);
    close(lock_fd);
    return;
  }

  if (_dump_disable_ts) {
    image_path = g_strconcat(_dump_dir, "/", "image.flx", NULL);
  } else {
    g_autoptr(GDateTime) dt = g_date_time_new_now_utc();
    g_autofree gchar *timestamp = g_date_time_format(dt, "%Y-%m-%dT%H:%M:%S");
    image_path = g_strconcat(_dump_dir, "/", timestamp, "-image.flx", NULL);
  }

  g_signal_emit_by_name(felixsrc, "take-flx-image", image_path, _dump_src_type,
                        _dump_context);
}

int main(int argc, char *argv[]) {
  g_autoptr(GError) error = NULL;
  g_autoptr(GOptionContext) optctx =
      g_option_context_new("- Multistream RTSP Server");
  g_option_context_add_main_entries(optctx, entries, NULL);
  g_option_context_add_group(optctx, gst_init_get_option_group());
  g_option_context_set_summary(optctx, MULTI_LAUNCH_SUMMARY);
  g_option_context_set_description(optctx, MULTI_LAUNCH_DESCRIPTION);
  if (!g_option_context_parse(optctx, &argc, &argv, &error)) {
    g_printerr("Error parsing options: %s\n", error->message);
    return -1;
  }

  GstElement *pipeline = gst_parse_launch(_src, &error);
  if (error) {
    g_printerr("Failed to parse src pipeline: %s\n", error->message);
    return -1;
  }

  if (gst_element_set_state(pipeline, GST_STATE_PLAYING) ==
      GST_STATE_CHANGE_FAILURE) {
    g_printerr("Failed to start source pipeline\n");
    return -1;
  }
  g_usleep(1e6);
  g_print("\n");

  GMainLoop *loop = g_main_loop_new(NULL, FALSE);
  GstRTSPServer *server = gst_rtsp_server_new();
  g_object_set(server, "service", _port, NULL);
  g_autoptr(GstRTSPMountPoints) mounts =
      gst_rtsp_server_get_mount_points(server);

  GVariant *sinks = g_variant_new_parsed(_sinks);
  GVariantIter iter;
  g_autofree gchar *path = NULL, *bin = NULL;
  g_variant_iter_init(&iter, sinks);
  while (g_variant_iter_loop(&iter, "{ss}", &path, &bin)) {
    GstRTSPMediaFactory *factory = gst_rtsp_media_factory_new();
    gst_rtsp_media_factory_set_launch(factory, bin);
    gst_rtsp_media_factory_set_shared(factory, TRUE);
    gst_rtsp_mount_points_add_factory(mounts, path, factory);
    g_print("Add RTSP mount point with URL rtsp://%s:%s%s\n", g_get_host_name(),
            _port, path);
  }

  if (_dump_src_type) {
    felixsrc = get_felixsrc_element(pipeline);
    if (!felixsrc) {
      g_printerr("Failed to find felixsrc/felixmultisrc GStreamer element. "
                 "Taking an image isn't supported\n");
      return -1;
    }
    g_signal_connect(felixsrc, "image-ready", (GCallback)image_ready_event,
                     NULL);
    signal(DUMP_FLX_SIGNAL, dump_to_flx);
  }

  gst_rtsp_server_attach(server, NULL);
  g_main_loop_run(loop);

  g_free(image_path);
  gst_object_unref(felixsrc);
  return 0;
}
